# freedsm-firmware

Project repository with all the needed files to compile Tasmota using the specific configuration/modifications of the FreeDSM photometer.

FreeDSM will send the data to your MQTT server (HomeAssistant, for example) if you configure it in the web interface (as any other Tasmota device) using messages FreeDSM. Additionally, from the Berry script `autoexec.bec`, the device will send the data to the G4S-DSM platform (and also debug data if a server is specified in .debug file)

# <a name="requirements"></a> Requirements

- [Visual Studio Code](https://code.visualstudio.com/)
- [PlatformIO extension for VSC](https://marketplace.visualstudio.com/items?itemName=platformio.platformio-ide)

Read the [G4S FreeDSM wiki](https://gitlab.citic.udc.es/lia2-publico/g4s/-/wikis/home) for the complete information about FreeDSM: components, assembly and flashing.

# <a name="development"></a> Development

## <a name="desarrollo-data"></a> Additional file (autoexec.bec, Berry script)

FreeDSM needs the autoexec.bec in the filesystem (inside `data` folder of the project). It will be packaged in the `firmware.factory.bin`. They are NOT included in `firmware.bin` for OTA upgrade (web), the filesystem is not updated and the `autoexec.bec` needs to be uploaded manually. In the [G4S FreeDSM wiki](https://gitlab.citic.udc.es/lia2-publico/g4s/-/wikis/home) the update process is described.

`autoexec.bec` manages display, buttons, magnitudes/luxes calculations and the data publishing to G4S-DSM server and to the debug data, if enabled.

### display.ini

`autoexec.bec` (Berry script) generates automaticallly the display.ini file for the Air 101 LCD (ST7735S oled, 0.9 inches). There are two versions manufactured and they are not automatically differentiable by software.

- If display shows white background, edit `display.ini` and replace the line `:i,21,20` with `:i,20,21`
- If display is distorted, edit the first line of `display.ini` replacing `161,83` with `160,80`

### .debug

Automatically generated by the Berry script including `"DEBUG_SERVER_URL":""` (no debug data published). You can add your own local debug server, for example:

    "DEBUG_SERVER_URL":"http://192.168.1.199"

The FreeDSM will send the data in JSON format to this IP address

### .calib

Automatically generated by the Berry script with generic calibration `"CalibCoefs":[0.36,0.99,0.0]`, used to calculate magnitudes calibrated (mag_calib).

### .calibat

Automatically generated by the Berry script with generic battery calibration `"CaliBat":[1950.0,2600.0]`, maximum and minimum battery level. Adjustable for your own battery (ANALOG1 value in the web interface)

# <a name="project"></a> Project load and compilation

In VisualStudio Code, menu File -> Open folder, select the project folder and VisualStudio will load the project and configure all.

## <a name="compile"></a> Compile project

Compile selecting `PlatformIO` -> group `Default` -> `Build`, as shown in the next capture:

![Build Firmware Screenshot](docs/screenshots_readme/build_firmware_tutorial.png)

This process will generate the firmware files in `.pio\build\tasmota32c3` folder

    build_output\firmware\firmware.factory.bin
    build_output\firmware\firmware.bin

## <a name="console-tasmota"></a> Specific FreeDSM commands in the Tasmota Console

    Sensor111 I     # Launch an instant reading (out of the 1 minute configured period)
    Sensor111 5     # Set integration time 500ms
    Sensor111 6     # Set integration time 600ms (default)
    Sensor111 H     # Set Gain to high
    Sensor111 M     # Set Gain to max (default)

## <a name="code-structure"></a> Code structure

The folder `tasmota` contains the code of the firmware. Here are described the added driver, sensor and some minor modification of the default Tasmota files.

folder **tasmota_xdrv_driver** (drivers container)

- Added `xdrv_111_freedsm_id` to generate the FreeDSM id

folder **tasmota_xsns_sensor** (sensors container)

- Added `xns_111_freedsm.ino`  (111 was a free number, but be careful if you try to upgrade to a newer version of Tasmota). It includes:

```
#ifdef USE_I2C
#ifdef USE_FREEDSM

#define XSNS_111 111
#define XI2C_40 40 // See I2CDEVICES.md

#define TSL2591_ADDRESS 0x29
```

`USE_I2C` and `USE_FREEDSM` allows the inclusion of the driver in the compiled binary. Driver I2C is number 40 and freedsm sensor 111

- FreeDSM needs the `xsns_63_aht1x.ino` (it covers AHT1x and AHT2x). AHT2x is needed for FreeDSM because AHT1x can not share the I2C bus with other devices (in our case, with TSL2591).

There are different board versions that include AHT20 (silver) and AHT21 (black). The calibration command in this file (Tasmota 13.1) is not correct (0xB1). It is important to modify it with this line (0xBE) to enable any of them in the different board versions:

```
#define AHTX_CMD 0xBE // 0xB1 only works for "AHT2x" labeled boards, 0xBE for all
```

folder **tasmota_support** (modification)

- The `support.ino` was modified to allow the use of pin 11 to ON/OFF the screen (removed the first part of IF inside FlashPin for ESP32C3).

file **user_config_override.h**

Here is the information about the version, the need of an `USER_MODULE`,the `USER_TEMPLATE` to configure the GPIOs of the board, the default URL for OTA (web interface upgrade), the MQTT teleperiod (every 60 seconds, get the measurements and send the MQTT message) and the number of TSL2591 measurements taken in every teleperiod.

```
#define FRIENDLY_NAME "FreeDSM"
#define FREEDSM_VERSION "X.X"
...
#define MODULE USER_MODULE
...
#define USER_TEMPLATE "{\"NAME\":\"FreeDSM ESP32C3\",\"GPIO\":[4704,1,736,704,32,33,800,768,34,35,1024,992,544,36,0,0,0,0,640,608,6210,0],\"FLAG\":0,\"BASE\":18}"
...
#define OTA_URL "https://gitlab.citic.udc.es/lia2-publico/g4s/-/wikis/fw/firmware.bin"
...
#define TELE_PERIOD 60
#define FREEDSM_TSL2591_SAMPLES 6
```

In the Tasmota console, you can launch "GPIO 255" to check the GPIOs configuration

In this file, you will also find the `I2CDRIVERS_?_??` lines to enable the different drivers (disabled the VEML6070 UV sensor because it has the same id than TSL2591, to avoid conflicts).

file **platform_override.ini** (in root, outside the tasmota folder)

Uncomment the line to enable the compilation for esp32c3

```
default_envs            = tasmota32c3
```

Select the board to configure 4MB flash (app 2880k and filesystem 320k) and include the Berry script "autoexec.bec" in the filesystem:

```
[env:tasmota32_base]
board                    = esp32_4M
custom_files_upload      = data/autoexec.bec
```


